#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>

#include "autxml.h"

#define INDENT_STRING "    "

int indent_level = 0;


#define AXIS_MAX      5
#define INTER_SPACE   200

int opt_positivate = 0;
int opt_dualize    = 0;

#define XML_OUTPUT   0
#define WS1S_OUTPUT  1
#define VATA_OUTPUT  2
#define DUMP_OUTPUT  -1

int opt_output = DUMP_OUTPUT;

char *file_name = NULL; 


Nfa mk_nfa(int vars_no, char **var_list, int state_no, int init_state, int acc_no, int *acc_list) {
  Nfa nfa = (Nfa) malloc(sizeof(struct nfa));

  nfa->vars_no = vars_no;

  nfa->var_type = (int *) malloc(vars_no * sizeof(int));

  int i;
  
  for (i = 0; i < vars_no; i ++) {
    if (islower(*(var_list[i])))
      nfa->var_type[i] = FOVAR;
    else if (isupper(*(var_list[i])))
      nfa->var_type[i] = MSOVAR;
    else {
      printf("illegal variable name: %s\n", var_list[i]);
      return 0;
    }
  }
  
  nfa->var_list = var_list;
  nfa->state_no = state_no;
  nfa->init_state = init_state;
  nfa->acc_no = acc_no;
  nfa->acc_list = acc_list;

  nfa->trans_no = 0;
  nfa->trans_list = (Trans *) malloc(MAXTRANS * sizeof(struct trans));

  return nfa;
}


void nfa_add_trans(Nfa nfa, int src, int dst, char *lab) {
  Trans trans = (Trans) malloc(sizeof(struct trans));

  trans->src = src;
  trans->dst = dst;

  trans->lab = (char *) malloc(strlen(lab));
  strcpy(trans->lab, lab);

  nfa->trans_list[nfa->trans_no ++] = trans;
}


int list_member(int len, int *list, int el) {
  int i;

  for (i = 0; i < len; i ++)
    if (list[i] == el)
      return 1;

  return 0;
}


void nfa_rem_trans(Nfa nfa, int src, int dst) {
  int i, list_no = 0;
  int *list = (int *) malloc(nfa->trans_no * sizeof(int));
  
  for (i = 0; i < nfa->trans_no; i ++) {
    Trans trans = nfa->trans_list[i];

    if (trans->src == src && trans->dst == dst)
      list[list_no ++] = i;
  }

  Trans *new_list = (Trans *) malloc((nfa->trans_no - list_no) * sizeof(Trans));
  int new_no = 0;
  
  for (i = 0; i < nfa->trans_no; i ++)
    if (!list_member(list_no, list, i))
      new_list[new_no ++] = nfa->trans_list[i];

  nfa->trans_no = new_no;
  nfa->trans_list = new_list;

  free(list);
}



void nfa_set_init(Nfa nfa, int init) {
  nfa->init_state = init;
}


void nfa_print_dump(Nfa nfa) {
  int i;
  
  printf("alphabet: ");
  
  for (i = 0; i < nfa->vars_no; i ++)
    printf("%s (%d)%s", nfa->var_list[i], nfa->var_type[i], i<nfa->vars_no-1 ? ", " : "\n");

  printf("%d states\n", nfa->state_no);
  printf("initial state: %d\n", nfa->init_state);
  printf("accepting state(s): ");

  for (i = 0; i < nfa->acc_no; i ++)
    printf("%d %s", nfa->acc_list[i], i<nfa->acc_no-1 ? ", " : "\n");

  for (i = 0; i < nfa->trans_no; i ++) {
    Trans trans = nfa->trans_list[i];
    printf("%d -- %s --> %d\n", trans->src, trans->lab, trans->dst);  
  }
}


void ws1s_declare_vars(int vars_no, char **var_list, int *var_type) {
}


void ws1s_print_cover(int state_no) {
}


void ws1s_print_init(int state_no, int init_state) {
}


void ws1s_print_final(int state_no, int acc_no, int *acc_list) {
}


void ws1s_print_transitions(Nfa nfa) {
}


void nfa_print_ws1s(Nfa nfa) {
  ws1s_declare_vars(nfa->vars_no, nfa->var_list, nfa->var_type);
  ws1s_print_cover(nfa->state_no);
  ws1s_print_init(nfa->state_no, nfa->init_state);
  ws1s_print_final(nfa->state_no, nfa->acc_no, nfa->acc_list);
  ws1s_print_transitions(nfa);
}


void nfa_pos(Nfa nfa) {
  int i;

  for (i = 0; i < nfa->trans_no; i ++) {
    Trans trans = nfa->trans_list[i];

    int j;

    for (j = 0; j < nfa->vars_no; j ++)
      if (nfa->var_type[j] == MSOVAR) {
	if (trans->lab[j] == '0')
	  trans->lab[j] = 'X';
      }
      else if (nfa->var_type[j] == FOVAR) {
	trans->lab[j] = 'X';
      }
  }
}


void nfa_dual(Nfa nfa) {
  int i;

  for (i = 0; i < nfa->trans_no; i ++) {
    Trans trans = nfa->trans_list[i];

    int j;

    for (j = 0; j < nfa->vars_no; j ++)
      if (nfa->var_type[j] == MSOVAR) {
	if (trans->lab[j] == '1')
	  trans->lab[j] = '0';
      }
  }
}


void output_noindent(char *fmt, ...) {
  va_list args;
  
  va_start(args, fmt);
  vprintf(fmt, args);
  va_end(args);
}


void output(char *fmt, ...) {
  va_list args;
  int i;
  
  for(i = 0; i < indent_level; i ++)
    printf(INDENT_STRING);
  
  va_start(args, fmt);
  vprintf(fmt, args);
  va_end(args);
}


void xml_prologue() {
  output("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n");
  output("<Structure label-on=\"Transition\" type=\"FiniteStateAutomaton\">\n");

  indent_level ++;
  
  output("<Name/>\n");
  output("<Description/>\n");
  output("<Formula/>\n");
}


void xml_alphabet(int vars_no, char **var_list) {
  output("<Alphabet type=\"Propositional\">\n");
  indent_level ++;
  
  int i;
  for (i = 0; i < vars_no; i ++)
    output("<Proposition>%s</Proposition>\n", var_list[i]);

  indent_level --;
  output("</Alphabet>\n");
}


void xml_states(int state_no, int init_state) {
  output("<StateSet>\n");
  indent_level ++;

  int i;

  for(i = 0; i < state_no; i ++) {
    int x = 100, y = 100;
    
    output("<State sid=\"%d\">\n", i);
    indent_level ++;
    
    output("<X>%d</X>\n", x + (i % AXIS_MAX) * INTER_SPACE);
    output("<Y>%d</Y>\n", y + (i / AXIS_MAX) * INTER_SPACE);
    output("<Properties/>\n");

    indent_level --;    
    output("</State>\n");
  }
  
  indent_level --;
  output("</StateSet>\n");
  output("<InitialStateSet>\n");
  indent_level ++;
  output("<StateID>%d</StateID>\n", init_state);
  indent_level --;
  output("</InitialStateSet>\n");
}


void xml_accepting(int acc_no, int *acc_list) {
  output("<Acc type=\"Classic\">\n");
  indent_level ++;

  int i;
  
  for (i = 0; i < acc_no; i ++)
    output("<StateID>%d</StateID>\n", acc_list[i]);
  
  indent_level --;
  output("</Acc>\n");
}


void xml_transitions(Nfa nfa) {
  output("<TransitionSet complete=\"false\">\n");
  indent_level ++;

  int i;
  
  for (i = 0; i < nfa->trans_no; i ++) {
    int j;
    Trans trans = nfa->trans_list[i];
    
    output("<Transition tid=\"%d\">\n", i);
    indent_level ++;
    output("<From>%d</From>\n", trans->src);
    output("<To>%d</To>\n", trans->dst);
    output("<Label>");

    for (j = 0; j < nfa->vars_no; j ++) {
      char c = trans->lab[j];

      if (c != 'X')       
	output_noindent("%s%s ", c == '0' ? "~" : "",
		      nfa->var_list[j]);
    }
    
    output_noindent("</Label>\n");
      
    output("<Properties/>\n");
    indent_level --;
    output("</Transition>\n");
  }
  
  indent_level --;
  output("</TransitionSet>\n");
}


void xml_epilogue() {
  output("<Properties/>\n");
  indent_level --;
  output("</Structure>\n");
} 


void nfa_print_xml(Nfa nfa) {
  xml_prologue();
  xml_alphabet(nfa->vars_no, nfa->var_list);
  xml_states(nfa->state_no, nfa->init_state);
  xml_transitions(nfa);
  xml_accepting(nfa->acc_no, nfa->acc_list);
  xml_epilogue();
}


void vata_prologue() {
  output("Automaton ");

  if (file_name)
    output("%s\n", file_name);
  else
    output("AAA\n");
}


void vata_states(int state_no) {
  output("States");

  int i;

  for (i = 0; i < state_no; i ++)
    output(" q%d", i);

  output("\n");
}


void vata_accepting(int acc_no, int *acc_list) {
  output("Final States");

  int i;

  for (i = 0; i < acc_no; i ++)
    output(" q%d", acc_list[i]);

  output("\n");
}


void vata_transitions(Nfa nfa) {
  output("Transitions\n");

  int i;

  for (i = 0; i < nfa->trans_no; i ++) {
    Trans trans = nfa->trans_list[i];

    int j;

    for (j = 0; j < MAXVAR - nfa->vars_no; j ++)
      output("X");
    
    for (j = 0; j < nfa->vars_no; j ++)
      output("%c", trans->lab[j]);

    if (trans->src == nfa->init_state)
      output(" -> q%d\n", trans->dst);
    else
      output("(q%d) -> q%d\n", trans->src, trans->dst);
  }
}


void nfa_print_vata(Nfa nfa) {
  vata_prologue();
  vata_states(nfa->state_no);
  vata_accepting(nfa->acc_no, nfa->acc_list);
  vata_transitions(nfa);
}

int parse_options(int argc, char **argv) {
  int pos = 1;

  while (pos < argc-1) {
    if (!strcmp(argv[pos], "-p"))
      opt_positivate = 1;
    else if (!strcmp(argv[pos], "-d"))
      opt_dualize = 1;
    else if (!strcmp(argv[pos], "-l")) {
      pos ++;
      if (!strcmp(argv[pos], "xml"))
	opt_output = XML_OUTPUT;
      else if (!strcmp(argv[pos], "ws1s"))
	opt_output = WS1S_OUTPUT;
      else if (!strcmp(argv[pos], "vata"))
	opt_output = VATA_OUTPUT;
      else if (!strcmp(argv[pos], "dump"))
	opt_output = DUMP_OUTPUT;
      else {
	printf("unknown output format: %s\n", argv[pos]);
	return 0;
      }
    }
    
    pos ++;
  }

  if (opt_dualize && !opt_positivate) {
    printf("option -d requires -p\n");
    return 0;
  }
  
  return pos;
}


int main(int argc, char **argv) {
  FILE *file;

  if (argc <= 1) {
    printf("usage: autlog <options> <file>\n");
    return 0;
  }

  int file_pos = parse_options(argc, argv);

  if (file_pos <= 0)
    return 0;

  if (file_pos >= argc) {
    printf("usage: autlog <options> <file>\n");
    return 0;    
  }
  
  file_name = argv[file_pos];
  
  file = fopen(file_name, "r");
  
  if (!file) {
    printf("%s: file not found\n", file_name);
    return 0;
  }

  /* read preamble */

  char *linebuf = (char *) malloc(LINESIZE);
  linebuf[0] = '\0';

  char **var_list = (char **) malloc(MAXVAR * sizeof(char *));
  char *var = (char *) malloc(LINESIZE);
  int var_no = 0;

  int start = strlen("DFA for formula with free variables: ");   
  
  while (strlen(linebuf) < start) {
    if (!fgets(linebuf, LINESIZE, file)) {
      printf("%s: error reading file\n", argv[1]);
      return 0;
    }
  }
  
  while (sscanf(linebuf+start, "%s ", var) == 1) {
    var_list[var_no] = (char *) malloc(strlen(var));
    strcpy(var_list[var_no ++], var);
    start += strlen(var) + 1;
  }

  if (!var_no) {
    printf("%s: the automaton has no variables\n", argv[1]);
    return 0;
  }
  
  int init_state;
  int *acc_list = (int *) malloc (MAXSTATES * sizeof(int));
  int acc_no = 0; 
  
  fscanf(file, "Initial state: %d\n", &init_state);
  fscanf(file, "Accepting states: ");
  
  while (fscanf(file, "%d ", &acc_list[acc_no]) == 1)
    acc_no ++;
  
  fscanf(file, "\n");

  int reject_state[MAXSTATES];
  int reject_no = 0;
  
  fscanf(file, "Rejecting states: ");

  while (fscanf(file, "%d ", &reject_state[reject_no ++]) == 1);

  fscanf(file, "\n");

  int dc;

  fscanf(file, "Don't-care states: ");

  while(fscanf(file, "%d ",  &dc) == 1);

  fscanf(file, "\n");
  
  int state_no, bdd_no;
  
  if (fscanf(file, "Automaton has %d states and %d BDD-nodes\n", &state_no, &bdd_no) != 2) {
    printf("%s: error reading file\n", argv[file_pos]);
    return 0;
  }

  if (init_state < 0 || init_state >= state_no) {
    printf("initial state %d out of range (%d)\n", init_state, state_no);
    return 0;
  }

  int i;
  
  for (i = 0; i < acc_no; i ++) {
    if (acc_list[i] < 0 || acc_list[i] >= state_no) {
      printf("accepting state %d out of range (%d)\n", acc_list[i], state_no);
      return 0;
    }
  }
  
  Nfa aut = mk_nfa(var_no, var_list, state_no, init_state, acc_no, acc_list);

  /* read transitions */

  fscanf(file, "Transitions:\n");

  int src, dst;
  char *label = (char *) malloc(var_no+1);

  while (fscanf(file, "State %d: %s -> state %d\n", &src, label, &dst) == 3) {
    if (src < 0 || src >= state_no) {
      printf("source state %d out of range (%d)\n", src, state_no);
      return 0;
    }

    if (dst < 0 || dst >= state_no) {
      printf("destination state %d out of range (%d)\n", dst, state_no);
      return 0;
    }    

    if (dst == init_state) {
      printf("\t*** WARNING: initial state %d is the destination of the transition %d -- %s --> %d\n",
	     init_state, src, label, dst);
      printf("\t*** translation to tree automata (vata) might not work correctly\n");
    }
    
    nfa_add_trans(aut, src, dst, label);
  }

  /* positivate and dualize transition labels */
  if (opt_positivate) {
    nfa_pos(aut);

    if (opt_dualize)
      nfa_dual(aut);
  }

  if (opt_output == XML_OUTPUT)
      nfa_print_xml(aut);
  else if (opt_output == WS1S_OUTPUT)
    nfa_print_ws1s(aut);
  else if (opt_output == VATA_OUTPUT)
    nfa_print_vata(aut);
  else if (opt_output == DUMP_OUTPUT)
    nfa_print_dump(aut);
  else {
    printf("unknown print format: \%d\n", opt_output);
    return 0;
  }
  
  return 1;
}
